import re
import typing as T
import snowfakery.data_gen_exceptions as exc
from dateutil.rrule import rrule, rruleset
from dateutil import rrule as rrule_mod
from datetime import datetime, date, time, timezone
from snowfakery import PluginResultIterator
from snowfakery.plugins import SnowfakeryPlugin, memorable
from snowfakery.template_funcs import parse_datetimespec, parse_date
from snowfakery.utils.validation_utils import resolve_value
from snowfakery.data_generator_runtime_object_model import SimpleValue, StructuredValue

# Note

#   As soon as the base rule and the exclusion differ by even a second,
#   the exclusion doesn't exclude properly. (I think microseconds are okay.)
#
#   E.g. if I exclude very Monday in December, but the "Monday" is Monday at 3:30
#        and the days are scheduled for 3:31, then the exclusion won't apply.

FREQ_STRS = {
    frequency: getattr(rrule_mod, frequency)
    for frequency in [
        "YEARLY",
        "MONTHLY",
        "WEEKLY",
        "DAILY",
        "HOURLY",
        "MINUTELY",
        "SECONDLY",
    ]
}


WEEKDAYS = {
    day: getattr(rrule_mod, day) for day in ["MO", "TU", "WE", "TH", "FR", "SA", "SU"]
}


def is_datetime(dt: str) -> bool:
    """Does a string look like a `datetime`?"""
    return bool(set(dt).intersection(" TZ+:"))


OptionalDateTimeLike = T.Union[str, date, datetime, None]
SeqOfIntsLike = T.Union[str, int, T.Sequence[int], None]
CalendarException = T.Union[
    T.Sequence[T.Union["CalendarRule", OptionalDateTimeLike]],
    "CalendarRule",
    OptionalDateTimeLike,
]


class CalendarRule(PluginResultIterator):
    """Generates calendar events"""

    iterator: T.Any  # Actually T.Union[T.Iterator[date],T.Iterator[ datetime]]

    def __init__(
        self,
        freq: str,
        start_date: OptionalDateTimeLike = None,
        interval: int = 1,
        count: T.Optional[int] = None,
        until: OptionalDateTimeLike = None,
        bysetpos: SeqOfIntsLike = None,  # undocumented feature, for now
        bymonth: SeqOfIntsLike = None,
        bymonthday: SeqOfIntsLike = None,
        byyearday: SeqOfIntsLike = None,
        byeaster: SeqOfIntsLike = None,  # undocumented feature for now
        byweekno: SeqOfIntsLike = None,  # undocumented feature for now
        byweekday: T.Optional[str] = None,
        byhour: SeqOfIntsLike = None,
        byminute: SeqOfIntsLike = None,
        bysecond: SeqOfIntsLike = None,
        cache: bool = False,  # undocumented feature for now
        exclude: CalendarException = None,
        include: CalendarException = None,
        use_undocumented_features: bool = False,
    ):
        # Certain features can only be used if you opt-in to use_undocumented_features
        self._check_undocumented_features(
            use_undocumented_features, bysetpos, byeaster, cache, byweekno
        )

        # Normalize and save the start_date
        self.start_date, precision = self._normalize_start_date(start_date)

        self._set_output_datetype_date_or_datetime(precision)

        wkst = rrule_mod.SU  # Standardize on Sunday as start of week globally
        # Could make this parameterizable if there is demand for it.
        bysetpos = process_list_of_ints(bysetpos)
        bymonth = process_list_of_ints(bymonth)
        bymonthday = process_list_of_ints(bymonthday)
        byyearday = process_list_of_ints(byyearday)
        byeaster = process_list_of_ints(byeaster)
        byhour = process_list_of_ints(byhour)
        byminute = process_list_of_ints(byminute)
        bysecond = process_list_of_ints(bysecond)
        byweekno = process_list_of_ints(bysecond)

        until = self._normalize_until(until)

        freq = self._normalize_frequency(freq)

        if byweekday:
            byweekday_normalized = self._normalize_weekday(byweekday)
        else:
            byweekday_normalized = None

        self.ruleset = rruleset(cache)
        self.rrule = rrule(
            freq=freq,
            dtstart=self.start_date,
            interval=interval,
            wkst=wkst,
            count=count,
            until=until,
            bysetpos=bysetpos,  # undocumented feature for now
            bymonth=bymonth,  # undocumented feature for now
            bymonthday=bymonthday,
            byyearday=byyearday,
            byeaster=byeaster,  # undocumented feature for now
            byweekno=byweekno,
            byweekday=byweekday_normalized,
            byhour=byhour,
            byminute=byminute,
            bysecond=bysecond,
            cache=cache,
        )
        self.ruleset.rrule(self.rrule)

        self.compound = include or exclude
        if exclude:
            self._process_special_cases(exclude, "exclude")

        if include:
            self._process_special_cases(include, "include")

        self.iterator = iter(self)

    def _check_undocumented_features(
        self,
        use_undocumented_features: bool,
        bysetpos: SeqOfIntsLike,
        byeaster: SeqOfIntsLike,
        cache: bool,
        byweekno: SeqOfIntsLike,
    ) -> None:
        """Require opt-in for undocumented features"""
        if (not use_undocumented_features) and any(
            [bysetpos, byeaster, cache, byweekno]
        ):
            raise exc.DataGenValueError(
                "That feature is undocumented and unsupported. "
                "Use the `use_undocumented_features: True` argument to use it regardless."
            )

    def _normalize_start_date(
        self, start_date: OptionalDateTimeLike
    ) -> T.Tuple[datetime, type]:
        """Normalize the start-date to datetime

        But return a type that keeps track of its original type"""

        if isinstance(start_date, str):
            precision = datetime if is_datetime(start_date) else date
            start_date = parse_datetimespec(start_date)
            assert start_date.tzinfo
        elif isinstance(start_date, datetime):
            precision = datetime
        elif isinstance(start_date, date):
            start_date = datetime.combine(
                start_date, time(hour=0, minute=0, second=0, tzinfo=timezone.utc)
            )
            precision = date
        elif not start_date:
            start_date = datetime.now()
            precision = datetime
        else:  # pragma: no cover
            raise TypeError(
                f"`start_date` ({start_date}) is of unknown type {type(start_date)}"
            )
        if not start_date.tzinfo:
            start_date = start_date.replace(tzinfo=timezone.utc)
        assert precision in (date, datetime)

        return start_date, precision

    def _normalize_until(self, until: OptionalDateTimeLike) -> T.Optional[datetime]:
        """Normalize until value to datetime or None"""
        if not until:
            return None

        if isinstance(until, str):
            if is_datetime(until):
                until = parse_datetimespec(until)
            else:
                until = datetime.combine(parse_date(until), self.start_date.time())

        elif isinstance(until, date):
            until = datetime.combine(until, self.start_date.time(), tzinfo=timezone.utc)

        else:
            raise exc.DataGenTypeError(
                f"`until` parameter ({until}) is of unexpected type {until}"
            )

        return until.replace(tzinfo=timezone.utc)

    def _set_output_datetype_date_or_datetime(self, precision: type) -> None:
        """Depending on the precision requested, generate the right kinds of records"""
        assert precision in (date, datetime)
        if precision is date:
            self.next = self._next_date  # type: ignore
        elif precision is datetime:
            self.next = self._next_datetime  # type: ignore
        else:  # pragma: no cover
            assert False, "Should be unreachable"

    def _process_special_cases(
        self,
        case: CalendarException,
        action: T.Literal["include", "exclude"],
    ) -> None:
        """Process inclusions and exclusions"""
        if action == "exclude":
            # https://dateutil.readthedocs.io/en/stable/rrule.html#dateutil.rrule.rruleset.exrule
            add_rule = self.ruleset.exrule
            # https://dateutil.readthedocs.io/en/stable/rrule.html#dateutil.rrule.rruleset.exdate
            add_date = self.ruleset.exdate
        elif action == "include":
            # https://dateutil.readthedocs.io/en/stable/rrule.html#dateutil.rrule.rruleset.rrule
            add_rule = self.ruleset.rrule
            # https://dateutil.readthedocs.io/en/stable/rrule.html#dateutil.rrule.rruleset.rdate
            add_date = self.ruleset.rdate
        else:  # pragma: no cover   -   Should be unreachable
            assert action in ("include", "exclude"), "Bad action!"
            raise NotImplementedError("Bad action!")

        if isinstance(case, (list, tuple)):
            for case in case:
                self._process_special_cases(case, action)
        elif isinstance(case, CalendarRule):
            add_rule(T.cast(T.Any, case.ruleset))

        elif isinstance(case, datetime):
            add_date(case)
        elif isinstance(case, date):
            d: date = case
            self._process_special_cases(
                datetime.combine(d, self.start_date.time(), tzinfo=timezone.utc), action
            )
        elif isinstance(case, str):
            d2: date = parse_date(case)
            dt: datetime = datetime.combine(
                d2, self.start_date.time(), tzinfo=timezone.utc
            )
            self._process_special_cases(dt, action)
        else:  # pragma: no cover
            raise TypeError(f"Cannot {action} {case}, ({type(case)})")

    def _normalize_frequency(self, freq: str):
        """Normalize frequency string to dateutil type"""
        try:
            freq_str = freq.upper()
            freq = FREQ_STRS[freq_str]
        except KeyError:
            raise exc.DataGenError(
                f"Cannot accept frequency string '{freq}'. Valid strings are {tuple(FREQ_STRS.keys())}"
            )

        except Exception as e:
            raise exc.DataGenError(
                f"Cannot accept frequency string '{freq}' because {repr(e)}"
            )
        if freq in (rrule_mod.HOURLY, rrule_mod.MINUTELY, rrule_mod.SECONDLY):

            if self.next != self._next_datetime:
                raise exc.DataGenError(
                    "If frequency is a time value, `start_date` should be a datetime"
                )

        return freq

    def _normalize_weekday(
        self, byweekday: T.Optional[str]
    ) -> T.List[rrule_mod.weekday]:
        """Normalize weekday string to dateutil types"""
        assert byweekday

        if isinstance(byweekday, str):
            weekdays = byweekday.split(",")
        else:
            raise exc.DataGenTypeError(
                f"Wrong type for byweekday {byweekday}, {type(byweekday)}"
            )

        days = [self._parse_weekday(day) for day in weekdays]

        return days

    def _parse_weekday(self, day):
        """Parse the weekday syntax like MO(+3) and WE(-2) and create dateutil datatypes"""
        try:
            if "(" in day:
                day, offset = self._split_weekday(day)
            else:
                offset = None
            dayconst = WEEKDAYS[day.upper().strip()]

        except (TypeError, KeyError):
            raise exc.DataGenTypeError(
                f"Do not know how to interpret {day} as a day of the week."
            )

        if offset:
            dayconst = dayconst(offset)
        return dayconst

    def _split_weekday(self, day) -> T.Tuple[str, int]:
        """Parse the weekday syntax like MO(+3) and WE(-2)"""
        parts = re.match(r"\s*(?P<weekday>\w+)\((?P<offset>.+)\)\s*", day)
        if not parts:
            raise exc.DataGenSyntaxError(f"Cannot understand {day}")
        weekday = parts["weekday"]
        offset = parts["offset"]
        try:
            offset = int(offset)
        except ValueError:
            raise exc.DataGenTypeError(f"Cannot interpret {offset} as a number")
        return weekday, offset

    def __iter__(self) -> T.Union[T.Iterator[datetime], T.Iterator[date]]:
        return iter(self.ruleset)

    def next(self) -> None:  # pragma: no cover
        """This method is never called.

        It is replaced at runtime by _next_datetime or _next_date"""
        raise NotImplementedError("next is not implemented")

    def _next_datetime(self) -> datetime:
        return next(self.iterator)

    def _next_date(self) -> date:
        val: datetime = next(self.iterator)
        return val.date()


class Schedule(SnowfakeryPlugin):
    class Functions:
        @memorable
        def Event(
            self,
            freq: str,
            start_date=None,
            interval=1,
            # wkst=None,
            count=None,
            until=None,
            bysetpos=None,  # undocumented feature for now
            bymonth=None,  # undocumented feature for now
            bymonthday=None,
            byyearday=None,
            byeaster=None,
            byweekno=None,
            byweekday=None,
            byhour=None,
            byminute=None,
            bysecond=None,
            cache=False,
            exclude=None,
            include=None,
        ):
            return CalendarRule(
                freq=freq,
                start_date=start_date,
                interval=interval,
                # wkst=wkst,
                count=count,
                until=until,
                bysetpos=bysetpos,  # No requirement identified yet
                bymonth=bymonth,  # No requirement id
                bymonthday=bymonthday,
                byyearday=byyearday,
                byeaster=byeaster,
                byweekno=byweekno,
                byweekday=byweekday,
                byhour=byhour,
                byminute=byminute,
                bysecond=bysecond,
                cache=cache,
                exclude=exclude,
                include=include,
            )

    class Validators:
        """Validators for Schedule plugin functions."""

        @staticmethod
        def validate_Event(sv, context):
            """Validate Schedule.Event(freq, ...) function call."""

            kwargs = getattr(sv, "kwargs", {})

            # ERROR: Required parameter 'freq'
            if "freq" not in kwargs:
                context.add_error(
                    "Schedule.Event: Missing required parameter 'freq'",
                    getattr(sv, "filename", None),
                    getattr(sv, "line_num", None),
                )
                return

            # Validate freq value
            freq_val = resolve_value(kwargs["freq"], context)

            if freq_val is not None:
                if not isinstance(freq_val, str):
                    context.add_error(
                        f"Schedule.Event: 'freq' must be a string, got {type(freq_val).__name__}",
                        getattr(sv, "filename", None),
                        getattr(sv, "line_num", None),
                    )
                else:
                    if freq_val.upper() not in FREQ_STRS:
                        context.add_error(
                            f"Schedule.Event: Invalid frequency '{freq_val}'. Valid values: {', '.join(FREQ_STRS.keys())}",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )

            # Validate start_date
            if "start_date" in kwargs:
                start_date_val = resolve_value(kwargs["start_date"], context)

                if start_date_val is not None and not isinstance(
                    start_date_val, (str, date, datetime)
                ):
                    context.add_error(
                        f"Schedule.Event: 'start_date' must be a string or date/datetime, got {type(start_date_val).__name__}",
                        getattr(sv, "filename", None),
                        getattr(sv, "line_num", None),
                    )

            # Validate interval
            if "interval" in kwargs:
                interval_val = resolve_value(kwargs["interval"], context)

                if interval_val is not None:
                    if not isinstance(interval_val, int):
                        context.add_error(
                            f"Schedule.Event: 'interval' must be an integer, got {type(interval_val).__name__}",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )
                    elif interval_val <= 0:
                        context.add_error(
                            "Schedule.Event: 'interval' must be positive",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )

            # Validate count
            if "count" in kwargs:
                count_val = resolve_value(kwargs["count"], context)

                if count_val is not None:
                    if not isinstance(count_val, int):
                        context.add_error(
                            f"Schedule.Event: 'count' must be an integer, got {type(count_val).__name__}",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )
                    elif count_val <= 0:
                        context.add_error(
                            "Schedule.Event: 'count' must be positive",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )

            # Validate until
            if "until" in kwargs:
                until_val = resolve_value(kwargs["until"], context)

                if until_val is not None and not isinstance(
                    until_val, (str, date, datetime)
                ):
                    context.add_error(
                        f"Schedule.Event: 'until' must be a string or date/datetime, got {type(until_val).__name__}",
                        getattr(sv, "filename", None),
                        getattr(sv, "line_num", None),
                    )

            # WARNING: Both count and until
            if "count" in kwargs and "until" in kwargs:
                context.add_warning(
                    "Schedule.Event: Using both 'count' and 'until' may produce unexpected results",
                    getattr(sv, "filename", None),
                    getattr(sv, "line_num", None),
                )

            # Validate byweekday
            if "byweekday" in kwargs:
                byweekday_val = resolve_value(kwargs["byweekday"], context)

                if byweekday_val is not None:
                    if not isinstance(byweekday_val, str):
                        context.add_error(
                            f"Schedule.Event: 'byweekday' must be a string, got {type(byweekday_val).__name__}",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )
                    else:
                        # Validate weekday format
                        days = [d.strip() for d in byweekday_val.split(",")]

                        for day in days:
                            # Extract day part (before parentheses if present)
                            day_part = day.split("(")[0].strip().upper()

                            if day_part not in WEEKDAYS:
                                context.add_error(
                                    f"Schedule.Event: Invalid weekday '{day_part}'. Valid values: {', '.join(WEEKDAYS.keys())}",
                                    getattr(sv, "filename", None),
                                    getattr(sv, "line_num", None),
                                )

            # Validate bymonthday
            Schedule.Validators._validate_day_list(
                sv, context, "bymonthday", min_val=1, max_val=31
            )

            # Validate byyearday
            Schedule.Validators._validate_day_list(
                sv, context, "byyearday", min_val=1, max_val=366
            )

            # Validate byhour
            Schedule.Validators._validate_time_component(sv, context, "byhour", 23)

            # Validate byminute
            Schedule.Validators._validate_time_component(sv, context, "byminute", 59)

            # Validate bysecond
            Schedule.Validators._validate_time_component(sv, context, "bysecond", 59)

            # Validate exclude
            Schedule.Validators._validate_exception(sv, context, "exclude")

            # Validate include
            Schedule.Validators._validate_exception(sv, context, "include")

            # WARNING: Unknown parameters
            valid_params = {
                "freq",
                "start_date",
                "interval",
                "count",
                "until",
                "bysetpos",
                "bymonth",
                "bymonthday",
                "byyearday",
                "byeaster",
                "byweekno",
                "byweekday",
                "byhour",
                "byminute",
                "bysecond",
                "cache",
                "exclude",
                "include",
                "use_undocumented_features",
                "parent",
                "_",
            }
            unknown = set(kwargs.keys()) - valid_params
            if unknown:
                context.add_warning(
                    f"Schedule.Event: Unknown parameter(s): {', '.join(sorted(unknown))}",
                    getattr(sv, "filename", None),
                    getattr(sv, "line_num", None),
                )

            # Return intelligent mock: date or datetime based on frequency
            # Try to resolve start_date if provided
            if "start_date" in kwargs:
                start_val = resolve_value(kwargs["start_date"], context)
                if isinstance(start_val, (date, datetime)):
                    return start_val
                elif isinstance(start_val, str):
                    try:
                        # Try to parse the start_date (imports are at module level)
                        if ":" in start_val or "T" in start_val:
                            return parse_datetimespec(start_val)
                        else:
                            return parse_date(start_val)
                    except Exception:
                        pass

            # Check frequency to determine return type
            freq_val = resolve_value(kwargs.get("freq"), context)
            if freq_val and isinstance(freq_val, str):
                freq_upper = freq_val.upper()
                if freq_upper in ("HOURLY", "MINUTELY", "SECONDLY"):
                    # Return datetime for time-based frequencies
                    return datetime.now(timezone.utc)

            # Default: return today's date
            return date.today()

        @staticmethod
        def _validate_day_list(sv, context, param_name, min_val=1, max_val=31):
            """Helper to validate day lists (bymonthday, byyearday, etc.)"""

            kwargs = getattr(sv, "kwargs", {})

            if param_name in kwargs:
                raw_val = kwargs[param_name]

                # Check raw type before resolution (to catch invalid types early)
                if not isinstance(raw_val, (int, str, SimpleValue)):
                    context.add_error(
                        f"Schedule.Event: '{param_name}' must be integer or string, got {type(raw_val).__name__}",
                        getattr(sv, "filename", None),
                        getattr(sv, "line_num", None),
                    )
                    return

                val = resolve_value(raw_val, context)

                if val is not None:
                    # Can be int or string
                    if isinstance(val, int):
                        vals = [val]
                    elif isinstance(val, str):
                        try:
                            vals = [int(v.strip()) for v in val.split(",")]
                        except ValueError:
                            context.add_error(
                                f"Schedule.Event: '{param_name}' must contain integers, got '{val}'",
                                getattr(sv, "filename", None),
                                getattr(sv, "line_num", None),
                            )
                            return
                    else:
                        context.add_error(
                            f"Schedule.Event: '{param_name}' must be integer or string, got {type(val).__name__}",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )
                        return

                    # Validate ranges
                    for v in vals:
                        if v == 0:
                            context.add_error(
                                f"Schedule.Event: '{param_name}' cannot be 0",
                                getattr(sv, "filename", None),
                                getattr(sv, "line_num", None),
                            )
                        elif v > 0 and (v < min_val or v > max_val):
                            context.add_error(
                                f"Schedule.Event: '{param_name}' must be between {min_val} and {max_val} (or negative), got {v}",
                                getattr(sv, "filename", None),
                                getattr(sv, "line_num", None),
                            )

        @staticmethod
        def _validate_time_component(sv, context, param_name, max_value):
            """Validate time components (hour, minute, second)"""

            kwargs = getattr(sv, "kwargs", {})

            if param_name in kwargs:
                raw_val = kwargs[param_name]

                # Check raw type before resolution (to catch invalid types early)
                if not isinstance(raw_val, (int, str, SimpleValue)):
                    context.add_error(
                        f"Schedule.Event: '{param_name}' must be integer or string, got {type(raw_val).__name__}",
                        getattr(sv, "filename", None),
                        getattr(sv, "line_num", None),
                    )
                    return

                val = resolve_value(raw_val, context)

                if val is not None:
                    # Can be int or comma-separated string
                    if isinstance(val, int):
                        vals = [val]
                    elif isinstance(val, str):
                        try:
                            vals = [int(v.strip()) for v in val.split(",")]
                        except ValueError:
                            context.add_error(
                                f"Schedule.Event: '{param_name}' must contain integers",
                                getattr(sv, "filename", None),
                                getattr(sv, "line_num", None),
                            )
                            return
                    else:
                        context.add_error(
                            f"Schedule.Event: '{param_name}' must be integer or string, got {type(val).__name__}",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )
                        return

                    # Validate range
                    for v in vals:
                        if v < 0 or v > max_value:
                            context.add_error(
                                f"Schedule.Event: '{param_name}' must be between 0 and {max_value}, got {v}",
                                getattr(sv, "filename", None),
                                getattr(sv, "line_num", None),
                            )

        @staticmethod
        def _validate_exception(sv, context, param_name):
            """Validate exclude/include parameters"""

            kwargs = getattr(sv, "kwargs", {})

            if param_name in kwargs:
                val = kwargs[param_name]

                # Can be: string (date), list, or StructuredValue (nested Schedule.Event)
                if isinstance(val, StructuredValue):
                    # It's a nested Schedule.Event - validate it recursively
                    if val.function_name != "Schedule.Event":
                        context.add_warning(
                            f"Schedule.Event: '{param_name}' expects Schedule.Event or date string, got {val.function_name}",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )
                elif isinstance(val, list):
                    # List of dates or nested events
                    for item in val:
                        # Recursively validate each item
                        if isinstance(item, StructuredValue):
                            if item.function_name != "Schedule.Event":
                                context.add_warning(
                                    f"Schedule.Event: '{param_name}' list item expects Schedule.Event or date string, got {item.function_name}",
                                    getattr(sv, "filename", None),
                                    getattr(sv, "line_num", None),
                                )
                        else:
                            # Validate date string
                            resolved_item = resolve_value(item, context)
                            if resolved_item is not None and not isinstance(
                                resolved_item, (str, date, datetime)
                            ):
                                context.add_error(
                                    f"Schedule.Event: '{param_name}' list item must be a date string or Schedule.Event, got {type(resolved_item).__name__}",
                                    getattr(sv, "filename", None),
                                    getattr(sv, "line_num", None),
                                )
                else:
                    # Should be a date string
                    resolved_val = resolve_value(val, context)
                    if resolved_val is not None and not isinstance(
                        resolved_val, (str, date, datetime)
                    ):
                        context.add_error(
                            f"Schedule.Event: '{param_name}' must be a date string, Schedule.Event, or list, got {type(resolved_val).__name__}",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )


def process_list_of_ints(val: SeqOfIntsLike) -> T.Optional[T.List[int]]:
    if val is None:
        return None
    elif isinstance(val, int):
        return [val]
    elif isinstance(val, str):
        return [int(v) for v in val.split(",")]
    elif isinstance(val, (list, tuple)):
        return [int(v) for v in val]
    else:
        raise exc.DataGenTypeError(
            f"Expected a number or list of numbers, not {val} ({type(val)})"
        )



"""Recipe validation framework.

This module provides semantic validation for Snowfakery recipes,
catching errors before runtime execution.
"""

import re
from typing import Dict, List, Optional, Any, Callable
from dataclasses import dataclass
from datetime import datetime, timezone
from dateutil.relativedelta import relativedelta
from faker import Faker
import jinja2
from jinja2 import nativetypes
from jinja2.sandbox import SandboxedEnvironment

from snowfakery.utils.validation_utils import (
    get_fuzzy_match,
    resolve_value,
    with_mock_context,
)
from snowfakery.data_generator_runtime_object_model import (
    ObjectTemplate,
    VariableDefinition,
    FieldFactory,
    ForEachVariableDefinition,
    StructuredValue,
    SimpleValue,
)
from snowfakery.plugins import PluginResultIterator
from snowfakery.template_funcs import StandardFuncs
from snowfakery.fakedata.faker_validators import FakerValidators
from snowfakery.fakedata.fake_data_generator import FakeNames
from snowfakery.utils.template_utils import StringGenerator
from snowfakery.object_rows import ObjectReference


class SandboxedNativeEnvironment(SandboxedEnvironment, nativetypes.NativeEnvironment):
    """Jinja2 environment that combines sandboxing security with native type preservation.

    This class provides:
    - Security restrictions from SandboxedEnvironment (blocks dangerous operations)
    - Native Python type preservation from NativeEnvironment (returns int, bool, list, etc.)

    Used during validation to safely execute Jinja templates while maintaining
    type compatibility with Snowfakery's runtime behavior.
    """

    def is_safe_attribute(self, obj, attr, value):
        """Override to allow access to __ prefixed temp variable fields on MockObjectRow.

        Snowfakery uses __ prefixed field names as temporary/hidden variables in recipes.
        These are legitimate field accesses, not attempts to access private Python attributes.
        """
        # Allow __ prefixed attributes on MockObjectRow objects (temp vars)
        if attr.startswith("__") and "MockObjectRow" in type(obj).__name__:
            return True
        # Fall back to default sandbox behavior for other cases
        return super().is_safe_attribute(obj, attr, value)


@dataclass
class ValidationError:
    """Represents a validation error."""

    message: str
    filename: Optional[str] = None
    line_num: Optional[int] = None

    def __str__(self):
        location = ""
        if self.filename:
            location = f"{self.filename}"
            if self.line_num:
                location += f":{self.line_num}"
            location += ": "
        return f"{location}Error: {self.message}"


@dataclass
class ValidationWarning:
    """Represents a validation warning."""

    message: str
    filename: Optional[str] = None
    line_num: Optional[int] = None

    def __str__(self):
        location = ""
        if self.filename:
            location = f"{self.filename}"
            if self.line_num:
                location += f":{self.line_num}"
            location += ": "
        return f"{location}Warning: {self.message}"


class ValidationResult:
    """Collects validation errors and warnings."""

    def __init__(
        self,
        errors: Optional[List[ValidationError]] = None,
        warnings: Optional[List[ValidationWarning]] = None,
    ):
        self.errors = errors if errors is not None else []
        self.warnings = warnings if warnings is not None else []

    def has_errors(self) -> bool:
        """Check if any errors were found."""
        return len(self.errors) > 0

    def has_warnings(self) -> bool:
        """Check if any warnings were found."""
        return len(self.warnings) > 0

    def get_summary(self) -> str:
        """Get a human-readable summary of validation results."""
        lines = []

        if self.errors:
            lines.append("\nValidation Errors:")
            for i, error in enumerate(self.errors, 1):
                lines.append(f"  {i}. {error.message}")
                if error.filename:
                    location = f"     at {error.filename}"
                    if error.line_num:
                        location += f":{error.line_num}"
                    lines.append(location)

        if self.warnings:
            lines.append("\nValidation Warnings:")
            for i, warning in enumerate(self.warnings, 1):
                lines.append(f"  {i}. {warning.message}")
                if warning.filename:
                    location = f"     at {warning.filename}"
                    if warning.line_num:
                        location += f":{warning.line_num}"
                    lines.append(location)

        if not self.errors and not self.warnings:
            lines.append("\n✓ Validation passed with no errors or warnings.")

        return "\n".join(lines)

    def __str__(self):
        return self.get_summary()


class ValidationContext:
    """Central context for validation with registries and error collection."""

    def __init__(self):
        # Function and provider registries
        self.available_functions: Dict[str, Callable] = {}
        self.faker_providers: set = set()

        # Dual object registries:
        # 1. ALL objects (pre-registered, for reference/random_reference validation)
        self.all_objects: Dict[str, Any] = {}  # All object names (forward refs allowed)
        self.all_nicknames: Dict[str, Any] = {}  # All nicknames (forward refs allowed)

        # 2. Sequential objects (registered as encountered, for Jinja ${{ObjectName.field}})
        self.available_objects: Dict[
            str, Any
        ] = {}  # Objects seen so far (order matters)
        self.available_nicknames: Dict[
            str, Any
        ] = {}  # Nicknames seen so far (order matters)

        # Variable registry (sequential, order matters)
        self.available_variables: Dict[
            str, Any
        ] = {}  # variable name -> VariableDefinition

        # Field registry within current object (for tracking field definition order)
        self.current_object_fields: Dict[
            str, Any
        ] = {}  # Fields defined so far in current object

        # Jinja environment (for syntax validation only)
        # Will be initialized in validate_recipe before any validation
        self.jinja_env: Any = None  # Jinja2 environment (jinja2.Environment)

        # Jinja execution support
        self.interpreter: Optional[Any] = None  # Interpreter instance
        self.current_template: Optional[
            Any
        ] = None  # Current ObjectTemplate/VariableDefinition being validated
        # Line number of current field being validated
        self.current_field_line_num: Optional[int] = None
        self.faker_instance: Optional[
            Any
        ] = None  # Faker instance for executing providers

        # Variable value cache to prevent recursion during evaluation
        self._variable_cache: Dict[str, Any] = {}
        self._evaluating: set = set()  # Track variables currently being evaluated

        # StructuredValue validation cache to prevent duplicate validation
        self._structured_value_cache: Dict[int, Any] = {}  # id(sv) -> mock result

        # Error collection
        self.errors: List[ValidationError] = []
        self.warnings: List[ValidationWarning] = []

    def add_error(
        self,
        message: str,
        filename: Optional[str] = None,
        line_num: Optional[int] = None,
    ):
        """Add a validation error."""
        self.errors.append(ValidationError(message, filename, line_num))

    def add_warning(
        self,
        message: str,
        filename: Optional[str] = None,
        line_num: Optional[int] = None,
    ):
        """Add a validation warning."""
        self.warnings.append(ValidationWarning(message, filename, line_num))

    def resolve_variable(self, name: str) -> Optional[Any]:
        """Look up a variable definition by name."""
        return self.available_variables.get(name)

    def resolve_object(
        self, name: str, allow_forward_ref: bool = False
    ) -> Optional[Any]:
        """Look up an object by name or nickname.

        Args:
            name: Object name or nickname to look up
            allow_forward_ref: If True, use all_objects (for reference/random_reference).
                              If False, use available_objects (for Jinja/sequential access).
        """
        if allow_forward_ref:
            # Use all_objects/all_nicknames (forward references allowed)
            if name in self.all_objects:
                return self.all_objects[name]
            if name in self.all_nicknames:
                return self.all_nicknames[name]
        else:
            # Use available_objects/available_nicknames (sequential, order matters)
            if name in self.available_objects:
                return self.available_objects[name]
            if name in self.available_nicknames:
                return self.available_nicknames[name]
        return None

    def get_object_count(self, obj_name: str) -> Optional[int]:
        """Get the literal count for an object if available."""
        obj = self.available_objects.get(obj_name)
        if obj and hasattr(obj, "count_expr"):
            # Try to extract literal count
            count_expr = obj.count_expr
            if isinstance(count_expr, int):
                return count_expr
            # For POC, we only handle literal integers
        return None

    def get_evaluator(self, definition: str):
        """Get Jinja evaluator (same as RuntimeContext).

        Args:
            definition: The Jinja template string

        Returns:
            A callable that evaluates the template, or a lambda returning the string if no Jinja
        """
        if not self.interpreter:
            raise RuntimeError("Interpreter not set in ValidationContext")
        return self.interpreter.template_evaluator_factory.get_evaluator(definition)

    def field_vars(self):
        """Build validation namespace for Jinja execution.

        Returns a dict mimicking the namespace available at runtime, but with mock values.
        This allows Jinja templates to be executed and validated.
        """
        return self._build_validation_namespace()

    def _build_validation_namespace(self):
        """Build namespace with mock values for all available names."""
        if not self.interpreter:
            raise RuntimeError("Interpreter not set in ValidationContext")

        namespace = {}

        # 1. Built-in variables (from EvaluationNamespace.simple_field_vars)
        namespace["id"] = 1
        namespace["count"] = 1
        namespace["child_index"] = 0
        namespace["today"] = self.interpreter.globals.today
        namespace["now"] = datetime.now(timezone.utc)
        namespace["template"] = self.current_template  # Current statement

        # Add 'this' - mock object representing current object being created
        if self.current_template and isinstance(self.current_template, ObjectTemplate):
            namespace["this"] = self._create_this_mock()
        else:
            namespace["this"] = None

        # 2. User variables (with mock values)
        for var_name in self.available_variables.keys():
            # Skip variables currently being evaluated to prevent recursion
            if var_name not in self._evaluating:
                namespace[var_name] = self._get_mock_value_for_variable(var_name)

        # 3. Objects (with field-aware mocks)
        for obj_name in self.available_objects.keys():
            namespace[obj_name] = self._create_mock_object(obj_name)
        for nickname in self.available_nicknames.keys():
            namespace[nickname] = self._create_mock_object(nickname)

        # 4. Functions (with validation wrappers)
        for func_name, validator in self.available_functions.items():
            namespace[func_name] = self._create_validation_function(
                func_name, validator
            )

        # 4.5. Constants from StandardFuncs (NULL, relativedelta, etc.)
        namespace["NULL"] = None
        namespace["null"] = None
        namespace["Null"] = None
        namespace["relativedelta"] = relativedelta

        # 4.6. Python builtins (available in Jinja at runtime)
        namespace["int"] = int
        namespace["str"] = str
        namespace["float"] = float
        namespace["bool"] = bool
        namespace["len"] = len
        namespace["list"] = list
        namespace["dict"] = dict
        namespace["set"] = set
        namespace["tuple"] = tuple
        namespace["min"] = min
        namespace["max"] = max
        namespace["sum"] = sum
        namespace["abs"] = abs
        namespace["round"] = round

        # 5. Plugins (with validation wrappers)
        for plugin_name, plugin_instance in self.interpreter.plugin_instances.items():
            namespace[plugin_name] = self._create_mock_plugin(
                plugin_name, plugin_instance
            )

        # 6. Faker (mock with provider validation)
        namespace["fake"] = self._create_mock_faker()

        # 7. Options
        namespace.update(self.interpreter.options)

        # 8. Current object fields (fields defined earlier in the same object)
        for field_name in self.current_object_fields.keys():
            if field_name not in self._evaluating:
                namespace[field_name] = self._get_mock_value_for_variable(field_name)

        return namespace

    def _get_mock_value_for_variable(self, var_name):
        """Get value for a variable or field.

        Args:
            var_name: Name of the variable or field

        Returns:
            The variable's evaluated value
        """
        # Check cache first
        if var_name in self._variable_cache:
            return self._variable_cache[var_name]

        # Mark as evaluating (to skip in namespace building and prevent recursion)
        self._evaluating.add(var_name)

        try:
            # Check available_variables first, then current_object_fields
            var_def = self.available_variables.get(var_name)
            if not var_def:
                var_def = self.current_object_fields.get(var_name)

            if var_def and hasattr(var_def, "expression"):
                expression = var_def.expression
            elif var_def and hasattr(var_def, "definition"):
                # For FieldFactory, use definition instead of expression
                expression = var_def.definition
            else:
                expression = None

            if expression:

                # If it's a SimpleValue, check if it's literal or Jinja
                if isinstance(expression, SimpleValue):
                    definition = expression.definition

                    # If it's a Jinja template, evaluate it
                    if isinstance(definition, str) and (
                        "${{" in definition or "${%" in definition
                    ):
                        result = validate_jinja_template_by_execution(
                            definition, expression.filename, expression.line_num, self
                        )
                        if result is not None:
                            self._variable_cache[var_name] = result
                            return result
                    else:
                        # Literal value
                        self._variable_cache[var_name] = definition
                        return definition

                # If it's a StructuredValue, resolve it
                if isinstance(expression, StructuredValue):
                    resolved = resolve_value(expression, self)
                    if resolved is not None:
                        # Plugins like Dataset and Counters return iterators
                        if isinstance(resolved, PluginResultIterator):
                            try:
                                resolved = resolved.next()
                            except StopIteration:
                                resolved = None

                        if resolved is not None:
                            if isinstance(resolved, ObjectReference):
                                ref_obj_name = resolved._tablename
                                ref_obj = self.resolve_object(
                                    ref_obj_name, allow_forward_ref=False
                                )
                                if ref_obj:
                                    resolved = self._create_mock_object(ref_obj_name)

                            self._variable_cache[var_name] = resolved
                            return resolved

            # Fall back to mock value if variable not found
            mock_value = f"<mock_variable_{var_name}>"
            self._variable_cache[var_name] = mock_value
            return mock_value
        finally:
            # Remove from evaluating set
            self._evaluating.discard(var_name)

    def _create_mock_object(self, name):
        """Create mock object that validates field access.

        Args:
            name: Object name or nickname

        Returns:
            MockObjectRow instance with field validation
        """
        is_this = name == "this"
        if is_this:
            obj_template = self.current_template
        else:
            obj_template = self.available_objects.get(
                name
            ) or self.available_nicknames.get(name)

        context = self

        class MockObjectRow:
            """Mock object that validates field access during validation."""

            def __init__(self, template, obj_name):
                self.id = 1
                self._child_index = 0
                self._template = template
                self._name = obj_name
                self._is_this = obj_name == "this"

                # Extract actual field names and definitions from template
                if template and hasattr(template, "fields"):
                    self._all_field_names = {
                        f.name for f in template.fields if isinstance(f, FieldFactory)
                    }
                    # Build field definition map
                    self._field_definitions = {
                        f.name: f.definition
                        for f in template.fields
                        if isinstance(f, FieldFactory)
                    }
                else:
                    self._all_field_names = set()
                    self._field_definitions = {}

            def __getattr__(self, attr):
                # Allow __ prefixed temp vars if they are valid fields
                # Only reject truly private attributes (single underscore internal attrs)
                # that are NOT in the accessible fields list
                if attr.startswith("_") and attr not in self._all_field_names:
                    raise AttributeError(f"'{attr}' not found")

                # For 'this': only fields defined so far are accessible
                # For object references: all fields of that object are accessible
                if self._is_this:
                    accessible_fields = set(context.current_object_fields.keys())
                else:
                    accessible_fields = self._all_field_names

                # Validate the attribute exists in accessible fields
                if attr not in accessible_fields:
                    display_name = (
                        "'this' object" if self._is_this else f"Object '{self._name}'"
                    )
                    raise AttributeError(
                        f"{display_name} has no field '{attr}'. "
                        f"Available fields: {', '.join(sorted(accessible_fields)) if accessible_fields else 'none'}"
                    )

                # Try to resolve the field value
                if attr in self._field_definitions:
                    from snowfakery.utils.validation_utils import resolve_value

                    field_def = self._field_definitions[attr]
                    try:
                        # Track error count before resolution to detect new errors
                        error_count_before = len(context.errors)
                        resolved = resolve_value(field_def, context)
                        # If resolution added errors (e.g., from context mismatch), remove them
                        # and fall back to mock value - these aren't real errors
                        if len(context.errors) > error_count_before:
                            # Remove errors added during this resolution attempt
                            context.errors = context.errors[:error_count_before]
                            return f"<mock_{self._name}.{attr}>"
                        if resolved is not None:
                            return resolved
                    except Exception:
                        # Fall back to mock value if resolution fails
                        pass

                # Fall back to mock value if we can't resolve
                return f"<mock_{self._name}.{attr}>"

        return MockObjectRow(obj_template, name)

    def _create_this_mock(self):
        """Create mock 'this' object for the current object being created."""
        return self._create_mock_object("this")

    def _create_validation_function(self, func_name, validator):
        """Create wrapper that validates when called from Jinja.

        Args:
            func_name: Name of the function
            validator: Validator function to call

        Returns:
            Wrapper function that validates and returns mock value
        """

        def validation_wrapper(*args, **kwargs):
            # Create synthetic StructuredValue
            # Use current_field_line_num if available (for inline Jinja calls), else use template line_num
            line_num = (
                self.current_field_line_num
                if self.current_field_line_num is not None
                else (self.current_template.line_num if self.current_template else 0)
            )
            sv = StructuredValue(
                func_name,
                list(args),
                self.current_template.filename
                if self.current_template
                else "<unknown>",
                line_num,
            )
            if kwargs:
                sv.kwargs = dict(kwargs)

            # Call validator and return its mock result
            try:
                return validator(sv, self)
            except Exception as exc:
                self.add_error(
                    f"Function '{func_name}' validation failed: {exc}",
                    sv.filename,
                    sv.line_num,
                )
                return f"<result_of_{func_name}>"

        return validation_wrapper

    def _create_mock_plugin(self, plugin_name, plugin_instance):
        """Create mock plugin namespace that validates function calls.

        Args:
            plugin_name: Name of the plugin (e.g., "StatisticalDistributions")
            plugin_instance: The actual plugin instance

        Returns:
            Mock plugin namespace with validated function wrappers
        """
        plugin_funcs = plugin_instance.custom_functions()

        class MockPlugin:
            def __init__(self, plugin_name, plugin_funcs, context):
                self._plugin_name = plugin_name
                self._plugin_funcs = plugin_funcs
                self._context = context

            def __getattr__(self, func_attr):
                # Build full function name with plugin namespace
                func_full_name = f"{self._plugin_name}.{func_attr}"

                # Check if this function has a validator
                if func_full_name in self._context.available_functions:
                    validator = self._context.available_functions[func_full_name]

                    # Create wrapper that calls validator and returns mock
                    def wrapper(*args, **kwargs):
                        # Use current_field_line_num if available (for inline Jinja calls), else use template line_num
                        line_num = (
                            self._context.current_field_line_num
                            if self._context.current_field_line_num is not None
                            else (
                                self._context.current_template.line_num
                                if self._context.current_template
                                else 0
                            )
                        )
                        sv = StructuredValue(
                            func_full_name,
                            kwargs if kwargs else list(args),
                            self._context.current_template.filename
                            if self._context.current_template
                            else "<unknown>",
                            line_num,
                        )

                        return validator(sv, self._context)

                    return wrapper
                else:
                    # No validator, return generic mock function
                    return lambda *args, **kwargs: f"<mock_{func_full_name}>"

        return MockPlugin(plugin_name, plugin_funcs, self)

    def _create_mock_faker(self):
        """Create mock Faker that validates provider names immediately.

        Returns:
            MockFaker instance that validates on attribute access
        """

        class MockFaker:
            def __init__(self, context):
                self.context = context

            def __getattr__(self, provider_name):
                # Get line number for error reporting
                line_num = (
                    self.context.current_field_line_num
                    if self.context.current_field_line_num is not None
                    else (
                        self.context.current_template.line_num
                        if self.context.current_template
                        else None
                    )
                )

                # Create StructuredValue for validate_fake (which validates provider name immediately)
                sv = StructuredValue(
                    "fake",
                    [provider_name],  # Just provider name, no args yet
                    self.context.current_template.filename
                    if self.context.current_template
                    else "<unknown>",
                    line_num,
                )

                # validate_fake returns a method - it validates the provider name immediately
                validated_method = FakerValidators.validate_fake(sv, self.context)

                # Wrap in StringGenerator for Jinja compatibility
                return StringGenerator(validated_method)

        return MockFaker(self)


def build_function_registry(plugins) -> Dict[str, Callable]:
    """Build registry mapping function names to validators.

    This maps actual function names (as they appear in recipes) to their validators.
    It handles cases where functions have Python-incompatible names (e.g., "if" -> "if_").

    Args:
        plugins: List of plugin objects that may contain Validators classes

    Returns:
        Dictionary mapping function names to their validator functions
    """
    registry = {}

    # Add standard function validators
    if hasattr(StandardFuncs, "Validators"):
        validators = StandardFuncs.Validators
        functions = (
            StandardFuncs.Functions if hasattr(StandardFuncs, "Functions") else None
        )

        for attr in dir(validators):
            if attr.startswith("validate_"):
                func_name = attr.replace("validate_", "")
                validator = getattr(validators, attr)
                registry[func_name] = validator

                # Check if there's an alias without trailing underscore (e.g., "if" for "if_")
                if functions and func_name.endswith("_"):
                    alias_name = func_name[:-1]
                    if hasattr(functions, alias_name):
                        # The Functions class has the alias (e.g., "if"), register it
                        registry[alias_name] = validator

    # Add plugin validators
    for plugin in plugins:
        plugin_name = plugin.__class__.__name__  # e.g., "UniqueId", "Math", etc.

        if hasattr(plugin, "Validators"):
            validators = plugin.Validators
            functions = plugin.Functions if hasattr(plugin, "Functions") else None

            for attr in dir(validators):
                if attr.startswith("validate_"):
                    func_name = attr.replace("validate_", "")
                    validator = getattr(validators, attr)

                    # Register with plugin namespace prefix for StructuredValue access
                    # e.g., "UniqueId.NumericIdGenerator"
                    registry[f"{plugin_name}.{func_name}"] = validator

                    # Check if there's an alias without trailing underscore
                    if functions and func_name.endswith("_"):
                        alias_name = func_name[:-1]
                        if hasattr(functions, alias_name):
                            registry[f"{plugin_name}.{alias_name}"] = validator

    # Add Faker validator (special case - StructuredValue syntax: fake: provider_name)
    registry["fake"] = FakerValidators.validate_fake

    return registry


def is_name_available(name: str, context: ValidationContext) -> bool:
    """Check if a name is available in the validation context.

    A name is considered available if it exists as:
    - A variable (in available_variables)
    - A function (in available_functions)
    - An object or nickname (in available_objects/available_nicknames)
    - A faker provider (in faker_providers)

    Args:
        name: The name to check
        context: The validation context

    Returns:
        True if the name is available, False otherwise
    """
    return (
        name in context.available_variables
        or name in context.available_functions
        or name in context.available_objects
        or name in context.available_nicknames
        or name in context.faker_providers
    )


def validate_recipe(parse_result, interpreter, options) -> ValidationResult:
    """Main entry point for recipe validation.

    Args:
        parse_result: The parsed recipe (ParseResult object)
        interpreter: Full Interpreter instance (with runtime context)
        options: User options passed to the recipe

    Returns:
        ValidationResult containing errors and warnings
    """
    # Build context
    context = ValidationContext()
    context.available_functions = build_function_registry(
        interpreter.plugin_instances.values()
    )

    # Store interpreter reference for Jinja execution
    context.interpreter = interpreter

    # Extract method names from faker by creating a Faker instance with the providers
    base_faker = Faker()

    # Add custom providers to the faker instance
    for provider in interpreter.faker_providers:
        base_faker.add_provider(provider)

    # Create a mock faker_context for FakeNames methods that need local_vars()
    class MockFakerContext:
        """Mock context for FakeNames during validation."""

        def local_vars(self):
            """Return empty dict (no previously generated fields during validation)."""
            return {}

    fake_names = FakeNames(base_faker, faker_context=MockFakerContext())

    # Create wrapper that combines base_faker and fake_names without circular reference
    class CombinedFaker:
        """Combines Faker and FakeNames, with FakeNames methods taking precedence."""

        def __init__(self, faker_inst, fake_names_inst):
            self._faker = faker_inst
            self._fake_names = fake_names_inst

        def __getattr__(self, name):
            # Check FakeNames first (takes precedence, like runtime)
            if hasattr(self._fake_names, name):
                return getattr(self._fake_names, name)
            # Fall back to faker
            return getattr(self._faker, name)

    faker_instance = CombinedFaker(base_faker, fake_names)

    # Store faker instance in context for execution
    context.faker_instance = faker_instance

    # Extract all callable methods from the faker instance
    faker_method_names = set()
    for name in dir(base_faker):
        if name.startswith("_"):
            continue
        try:
            attr = getattr(base_faker, name, None)
            if callable(attr):
                faker_method_names.add(name)
        except (TypeError, AttributeError):
            continue
    for name in dir(fake_names):
        if name.startswith("_"):
            continue
        try:
            attr = getattr(fake_names, name, None)
            if callable(attr):
                faker_method_names.add(name)
        except (TypeError, AttributeError):
            continue
    context.faker_providers = faker_method_names

    # Create Jinja environment with SandboxedNativeEnvironment
    context.jinja_env = SandboxedNativeEnvironment(
        block_start_string="${%",
        block_end_string="%}",
        variable_start_string="${{",
        variable_end_string="}}",
        undefined=jinja2.StrictUndefined,
    )

    def register_to_all(obj_template):
        """Register an ObjectTemplate to all_objects/all_nicknames."""
        context.all_objects[obj_template.tablename] = obj_template
        if obj_template.nickname:
            context.all_nicknames[obj_template.nickname] = obj_template

    # First pass: Pre-register ALL objects in all_objects/all_nicknames
    # This allows forward references for reference functions
    for statement in parse_result.statements:
        _walk_nested_objects(statement, register_to_all)

    # Second pass: Sequential validation with progressive registration
    for statement in parse_result.statements:
        # Set current template for Jinja context
        context.current_template = statement

        # Validate statement
        validate_statement(statement, context)

        # Clear current template
        context.current_template = None

        # Register variables (ObjectTemplates are registered inside validate_statement)
        if isinstance(statement, VariableDefinition):
            # Register variable (order matters for variables)
            context.available_variables[statement.varname] = statement

    return ValidationResult(context.errors, context.warnings)


def _walk_nested_objects(obj, on_object_found, visited=None):
    """Walk through nested structures and call callback for each ObjectTemplate found.

    Args:
        obj: Object to walk (ObjectTemplate, StructuredValue, list, dict, etc.)
        on_object_found: Callback function(obj_template) called for each ObjectTemplate
        visited: Set of visited object IDs to prevent infinite loops
    """
    if visited is None:
        visited = set()

    if isinstance(obj, ObjectTemplate):
        obj_id = id(obj)
        if obj_id in visited:
            return
        visited.add(obj_id)

        # Call callback for this object
        on_object_found(obj)

        # Recursively walk friends and fields
        for friend in obj.friends:
            _walk_nested_objects(friend, on_object_found, visited)
        for field in obj.fields:
            if isinstance(field, FieldFactory):
                _walk_nested_objects(field.definition, on_object_found, visited)

    elif isinstance(obj, StructuredValue):
        args = getattr(obj, "args", [])
        if isinstance(args, (list, tuple)):
            for arg in args:
                _walk_nested_objects(arg, on_object_found, visited)
        elif args is not None:
            _walk_nested_objects(args, on_object_found, visited)

        kwargs = getattr(obj, "kwargs", {})
        if isinstance(kwargs, dict):
            for value in kwargs.values():
                _walk_nested_objects(value, on_object_found, visited)

    elif isinstance(obj, FieldFactory):
        _walk_nested_objects(obj.definition, on_object_found, visited)

    elif isinstance(obj, (list, tuple)):
        for item in obj:
            _walk_nested_objects(item, on_object_found, visited)

    elif isinstance(obj, dict):
        for value in obj.values():
            _walk_nested_objects(value, on_object_found, visited)


def validate_statement(statement, context: ValidationContext):
    """Validate a single statement (ObjectTemplate or VariableDefinition).

    Args:
        statement: An ObjectTemplate or VariableDefinition
        context: The validation context
    """
    if isinstance(statement, ObjectTemplate):
        # Clear field registry for new object
        context.current_object_fields = {}

        # Validate count expression
        if statement.count_expr:
            validate_field_definition(statement.count_expr, context)

        # Validate and register for_each expression
        if statement.for_each_expr:
            if isinstance(statement.for_each_expr, ForEachVariableDefinition):
                # Validate the iterator expression
                validate_field_definition(statement.for_each_expr.expression, context)

                # Register the loop variable so fields can reference it
                context.available_variables[
                    statement.for_each_expr.varname
                ] = statement.for_each_expr

        # Helper to register nested objects to available registries
        def register_to_available(obj_template):
            context.available_objects[obj_template.tablename] = obj_template
            if obj_template.nickname:
                context.available_nicknames[obj_template.nickname] = obj_template

        # Validate fields sequentially (order matters within object)
        for field in statement.fields:
            if isinstance(field, FieldFactory):
                # Validate field (can reference previously defined fields in this object)
                validate_field_definition(field.definition, context)

                # Register any nested objects found in the field definition
                _walk_nested_objects(field.definition, register_to_available)

                # Register field so subsequent fields can reference it
                context.current_object_fields[field.name] = field

        # Register parent object AFTER validating fields but BEFORE validating friends
        context.available_objects[statement.tablename] = statement
        if statement.nickname:
            context.available_nicknames[statement.nickname] = statement

        # Recursively validate friends (nested ObjectTemplates)
        for friend in statement.friends:
            if isinstance(friend, ObjectTemplate):
                # Save and set current_template for nested validation
                saved_template = context.current_template
                context.current_template = friend

                # Validate the friend
                validate_statement(friend, context)

                # Restore current_template
                context.current_template = saved_template

                # Register friend in sequential registries AFTER validating
                # This ensures a friend can't random_reference itself on first instance
                context.available_objects[friend.tablename] = friend
                if friend.nickname:
                    context.available_nicknames[friend.nickname] = friend

    elif isinstance(statement, VariableDefinition):
        validate_field_definition(statement.expression, context)


def validate_jinja_template_by_execution(
    template_str: str, filename: str, line_num: int, context: ValidationContext
) -> Optional[Any]:
    """Validate Jinja template by executing it in validation context.

    This function actually executes the Jinja template in a mock context,
    catching any errors that would occur at runtime.

    Args:
        template_str: The Jinja template string
        filename: Source file for error reporting
        line_num: Line number for error reporting
        context: Validation context

    Returns:
        The resolved value if execution succeeds, None if it fails
    """
    # 1. Syntax checks
    try:
        context.jinja_env.parse(template_str)
    except jinja2.TemplateSyntaxError as e:
        context.add_error(f"Jinja syntax error: {str(e)}", filename, line_num)
        return None

    # 2. Check if template contains Jinja
    if not ("${{" in template_str or "${%" in template_str):
        # No Jinja template, just return the literal string
        return template_str

    # 3. Parse and execute template using our strict Jinja environment
    try:
        # Store the field's line number for inline function calls
        saved_line_num = context.current_field_line_num
        context.current_field_line_num = line_num

        try:
            namespace_dict = {}
            with with_mock_context(context, namespace_dict):
                namespace = context.field_vars()
                # Render the template (mock context is still active)
                template = context.jinja_env.from_string(template_str)
                result = template.render(namespace)
                # NativeEnvironment returns a lazy object - force evaluation to catch errors
                bool(result)  # Force evaluation
                return result
        finally:
            # Restore the previous line number
            context.current_field_line_num = saved_line_num
    except jinja2.exceptions.UndefinedError as e:
        # Variable or name not found
        error_msg = getattr(e, "message", str(e))

        # Simplify error messages about Mock* objects to be more user-friendly
        if error_msg and "has no attribute" in error_msg:
            match = re.search(r"has no attribute '(\w+)'", error_msg)
            if match:
                attr_name = match.group(1)
                if "MockObjectRow object" in error_msg:
                    error_msg = f"Object has no attribute '{attr_name}'"
                elif "MockPlugin object" in error_msg:
                    error_msg = f"Plugin has no attribute '{attr_name}'"

        context.add_error(
            f"Jinja template error: {error_msg}",
            filename,
            line_num,
        )
        return None
    except AttributeError as e:
        # Attribute access error (e.g., object.nonexistent_field)
        context.add_error(
            f"Jinja template attribute error: {str(e)}", filename, line_num
        )
        return None
    except TypeError as e:
        # Type error (e.g., calling non-callable, wrong arguments)
        context.add_error(f"Jinja template type error: {str(e)}", filename, line_num)
        return None
    except Exception as e:
        # Any other runtime error
        context.add_error(
            f"Jinja template evaluation error: {str(e)}", filename, line_num
        )
        return None


def validate_field_definition(field_def, context: ValidationContext):
    """Validate a FieldDefinition (SimpleValue or StructuredValue).

    Args:
        field_def: A FieldDefinition object (SimpleValue or StructuredValue)
        context: The validation context

    Returns:
        For StructuredValue: The mock result returned by the validator (or None)
        For SimpleValue: The resolved value (or None)
    """
    # Check if it's a StructuredValue (function call)
    if isinstance(field_def, StructuredValue):
        # Check if this StructuredValue has already been validated (using object id)
        sv_id = id(field_def)
        if sv_id in context._structured_value_cache:
            return context._structured_value_cache[sv_id]

        func_name = field_def.function_name
        mock_result = None

        # STEP 1: Resolve nested StructuredValues in args/kwargs BEFORE calling validator
        # This allows validators to receive mock values instead of StructuredValue objects
        resolved_args = []
        for arg in field_def.args:
            if isinstance(arg, StructuredValue):
                nested_mock = validate_field_definition(arg, context)
                resolved_args.append(nested_mock)
            else:
                resolved_args.append(arg)

        resolved_kwargs = {}
        for key, value in field_def.kwargs.items():
            if isinstance(value, StructuredValue):
                nested_mock = validate_field_definition(value, context)
                resolved_kwargs[key] = nested_mock
            else:
                resolved_kwargs[key] = value

        func_name = field_def.function_name
        lookup_func_name = func_name
        fake_provider = None
        if lookup_func_name not in context.available_functions and "." in func_name:
            base_name, method_name = func_name.split(".", 1)
            if base_name == "fake":
                lookup_func_name = "fake"
                fake_provider = method_name

        # STEP 2: Temporarily replace args/kwargs (and possibly function name) with resolved versions
        original_args = field_def.args
        original_kwargs = field_def.kwargs
        original_function_name = field_def.function_name

        if fake_provider:
            resolved_args = [fake_provider] + resolved_args
            field_def.function_name = "fake"

        field_def.args = resolved_args
        field_def.kwargs = resolved_kwargs

        try:
            # STEP 3: Look up validator and call it with resolved args/kwargs
            if lookup_func_name in context.available_functions:
                validator = context.available_functions[lookup_func_name]
                try:
                    result = validator(field_def, context)

                    # STEP 3.5: If validator returned a callable (like validate_fake does),
                    # call it with the resolved args from the StructuredValue
                    if callable(result) and lookup_func_name == "fake":
                        # For fake, args[0] is provider name, args[1:] are faker arguments
                        faker_args = resolved_args[1:] if len(resolved_args) > 1 else []
                        faker_kwargs = resolved_kwargs
                        mock_result = result(*faker_args, **faker_kwargs)
                    else:
                        mock_result = result
                except Exception as e:
                    # Catch any validator errors to avoid breaking the validation process
                    context.add_error(
                        f"Internal validation error for '{func_name}': {str(e)}",
                        getattr(field_def, "filename", None),
                        getattr(field_def, "line_num", None),
                    )
            else:
                # Check if it's a plugin function without a validator
                if "." in func_name:
                    plugin_name, method_name = func_name.split(".", 1)
                    if plugin_name in context.interpreter.plugin_instances:
                        # Plugin exists but function has no validator - return generic mock
                        mock_result = f"<result_of_{func_name}>"
                    else:
                        # Plugin doesn't exist - report error
                        suggestion = get_fuzzy_match(
                            func_name, list(context.available_functions.keys())
                        )
                        msg = f"Unknown function '{func_name}'"
                        if suggestion:
                            msg += f". Did you mean '{suggestion}'?"
                        context.add_error(
                            msg,
                            getattr(field_def, "filename", None),
                            getattr(field_def, "line_num", None),
                        )
                else:
                    # Unknown function - add error with suggestion
                    suggestion = get_fuzzy_match(
                        func_name, list(context.available_functions.keys())
                    )
                    msg = f"Unknown function '{func_name}'"
                    if suggestion:
                        msg += f". Did you mean '{suggestion}'?"
                    context.add_error(
                        msg,
                        getattr(field_def, "filename", None),
                        getattr(field_def, "line_num", None),
                    )
        finally:
            # STEP 4: Restore original args/kwargs
            field_def.args = original_args
            field_def.kwargs = original_kwargs
            field_def.function_name = original_function_name

        # Cache the result to prevent duplicate validation
        context._structured_value_cache[sv_id] = mock_result
        return mock_result

    # Check if it's a SimpleValue (literal or Jinja template)
    elif isinstance(field_def, SimpleValue):
        if isinstance(field_def.definition, str) and "${{" in field_def.definition:
            # It's a Jinja template - validate it
            result = validate_jinja_template_by_execution(
                field_def.definition, field_def.filename, field_def.line_num, context
            )
            return result
        else:
            # Return the literal value
            return field_def.definition if hasattr(field_def, "definition") else None

    return None



from pathlib import Path

from snowfakery.plugins import SnowfakeryPlugin
from snowfakery.utils.validation_utils import resolve_value


class File(SnowfakeryPlugin):
    class Functions:
        def file_data(self, file, encoding="utf-8"):
            if encoding == "binary":
                encoding = "latin-1"

            template_path = Path(self.context.current_filename).parent

            with open(template_path / file, "rb") as data:
                return data.read().decode(encoding)

    class Validators:
        """Validators for File plugin functions."""

        @staticmethod
        def validate_file_data(sv, context):
            """Validate File.file_data(file, encoding="utf-8")

            Args:
                sv: StructuredValue with args/kwargs
                context: ValidationContext for error reporting

            Returns:
                str: Mock file content or actual file content if file exists
            """
            kwargs = getattr(sv, "kwargs", {})
            args = getattr(sv, "args", [])

            # Check if file is provided (as positional or keyword argument)
            has_file = len(args) > 0 or "file" in kwargs

            if not has_file:
                context.add_error(
                    "File.file_data: Missing required parameter 'file'",
                    getattr(sv, "filename", None),
                    getattr(sv, "line_num", None),
                )
                return "Mock file content"  # Fallback mock

            # Validate file parameter
            if args:
                file_val = resolve_value(args[0], context)
            else:
                file_val = resolve_value(kwargs.get("file"), context)

            if file_val is not None:
                # ERROR: Must be string
                if not isinstance(file_val, str):
                    context.add_error(
                        f"File.file_data: 'file' must be a string, got {type(file_val).__name__}",
                        getattr(sv, "filename", None),
                        getattr(sv, "line_num", None),
                    )
                else:
                    # ERROR: File must exist
                    # Get the recipe file's directory
                    if context.current_template and context.current_template.filename:
                        template_path = Path(context.current_template.filename).parent
                        file_path = template_path / file_val

                        if not file_path.exists():
                            context.add_error(
                                f"File.file_data: File '{file_val}' does not exist (resolved to: {file_path})",
                                getattr(sv, "filename", None),
                                getattr(sv, "line_num", None),
                            )
                        elif not file_path.is_file():
                            context.add_error(
                                f"File.file_data: Path '{file_val}' exists but is not a file (resolved to: {file_path})",
                                getattr(sv, "filename", None),
                                getattr(sv, "line_num", None),
                            )

            # Validate encoding parameter (optional)
            if "encoding" in kwargs:
                encoding_val = resolve_value(kwargs["encoding"], context)

                if encoding_val is not None:
                    # ERROR: Must be string
                    if not isinstance(encoding_val, str):
                        context.add_error(
                            f"File.file_data: 'encoding' must be a string, got {type(encoding_val).__name__}",
                            getattr(sv, "filename", None),
                            getattr(sv, "line_num", None),
                        )

            # WARNING: Unknown parameters
            valid_params = {"file", "encoding"}
            unknown = set(kwargs.keys()) - valid_params
            if unknown:
                context.add_warning(
                    f"File.file_data: Unknown parameter(s): {', '.join(sorted(unknown))}",
                    getattr(sv, "filename", None),
                    getattr(sv, "line_num", None),
                )

            # Return intelligent mock: mock file content string
            return "Mock file content"



"""Validators for Faker provider calls using introspection."""

import inspect
from typing import get_origin, get_args, Union, Optional, Literal
from snowfakery.utils.validation_utils import resolve_value, get_fuzzy_match


class FakerValidators:
    """Validates Faker provider calls using introspection.

    This class uses Python's inspect module to introspect Faker method signatures
    and validate parameter names, counts, and types (when type annotations are available).
    """

    def __init__(self, faker_instance, faker_providers=None):
        """Initialize validator with Faker instance.

        Args:
            faker_instance: The Faker instance to introspect
            faker_providers: Set of available provider names (optional, extracted from faker_instance if not provided)
        """
        self.faker_instance = faker_instance
        self.faker_providers = faker_providers or self._extract_providers()
        self._signature_cache = {}  # Cache signatures for performance

    def _extract_providers(self):
        """Extract provider names from faker instance."""
        if not self.faker_instance:
            return set()

        providers = set()
        skip_attrs = {
            "seed",
            "seed_instance",
            "seed_locale",
        }  # Methods that raise errors
        for name in dir(self.faker_instance):
            if not name.startswith("_") and name not in skip_attrs:
                try:
                    attr = getattr(self.faker_instance, name)
                    if callable(attr):
                        providers.add(name)
                except (TypeError, AttributeError):
                    # Skip attributes that raise errors on access
                    pass
        return providers

    def _resolve_provider_name(self, provider_name: str) -> Optional[str]:
        """Resolve provider name accounting for case and underscores."""
        if not provider_name:
            return None
        if provider_name in self.faker_providers:
            return provider_name

        normalized = provider_name.replace("_", "").lower()
        for candidate in self.faker_providers:
            if candidate.lower() == provider_name.lower():
                return candidate
            if candidate.replace("_", "").lower() == normalized:
                return candidate
        return None

    def validate_provider_name(
        self, provider_name, context, filename=None, line_num=None
    ):
        """Validate that a provider name exists.

        Args:
            provider_name: Name of the Faker provider
            context: ValidationContext for error reporting
            filename: Source filename for error reporting
            line_num: Line number for error reporting

        Returns:
            Resolved provider name string if provider exists, otherwise None
        """
        resolved_name = self._resolve_provider_name(provider_name)
        if not resolved_name:
            suggestion = get_fuzzy_match(provider_name, list(self.faker_providers))
            msg = f"Unknown Faker provider '{provider_name}'"
            if suggestion:
                msg += f". Did you mean '{suggestion}'?"
            context.add_error(msg, filename, line_num)
            return None
        return resolved_name

    def validate_provider_call(self, provider_name, args, kwargs, context):
        """Validate a Faker provider call.

        This method validates:
        - Parameter names (catches typos)
        - Parameter counts (too many/few arguments)
        - Parameter types (when type annotations available)

        Args:
            provider_name: Name of the Faker provider (e.g., "email", "first_name")
            args: Positional arguments (list)
            kwargs: Keyword arguments (dict)
            context: ValidationContext for error reporting
        """
        resolved_name: Optional[str] = self._resolve_provider_name(provider_name)
        if not resolved_name:
            # Re-use name validation to record a helpful error message
            self.validate_provider_name(provider_name, context)
            return

        # 1. Check if provider exists
        if not hasattr(self.faker_instance, resolved_name):
            return

        # 2. Get the method
        method = getattr(self.faker_instance, resolved_name)

        # 3. Get signature (with caching)
        if resolved_name not in self._signature_cache:
            try:
                sig = inspect.signature(method)
                self._signature_cache[resolved_name] = sig
            except (ValueError, TypeError):
                # Can't introspect (rare case) - skip validation
                return

        sig = self._signature_cache[resolved_name]

        # 4. Resolve arguments (convert FieldDefinitions to actual values)
        resolved_args = []
        for arg in args:
            resolved = resolve_value(arg, context)
            # Use resolved value if available, otherwise use original
            resolved_args.append(resolved if resolved is not None else arg)

        resolved_kwargs = {}
        for key, value in kwargs.items():
            resolved = resolve_value(value, context)
            resolved_kwargs[key] = resolved if resolved is not None else value

        # 5. Validate parameter names and counts using sig.bind()
        try:
            bound = sig.bind(*resolved_args, **resolved_kwargs)
        except TypeError as e:
            # Parameter validation failed (wrong names, counts, etc.)
            filename = (
                context.current_template.filename if context.current_template else None
            )
            line_num = (
                context.current_template.line_num if context.current_template else None
            )
            context.add_error(
                f"fake.{provider_name}: {str(e)}",
                filename,
                line_num,
            )
            return

        # 6. Type checking (if parameters have type annotations)
        explicitly_provided = set(bound.arguments.keys())
        bound.apply_defaults()
        for param_name, param_value in bound.arguments.items():
            param_obj = sig.parameters[param_name]

            # Skip if no annotation
            if param_obj.annotation == inspect.Parameter.empty:
                continue

            # Skip type checking for default values - only check explicitly provided args
            if param_name not in explicitly_provided:
                continue

            # Only validate if we have a resolved literal value
            if not isinstance(param_value, (int, float, str, bool, type(None))):
                # Can't validate non-literal values (complex expressions)
                continue

            # Check type compatibility
            expected_type = param_obj.annotation
            if not self._check_type(param_value, expected_type):
                # Type mismatch - report error
                filename = (
                    context.current_template.filename
                    if context.current_template
                    else None
                )
                line_num = (
                    context.current_template.line_num
                    if context.current_template
                    else None
                )

                context.add_error(
                    f"fake.{provider_name}: Parameter '{param_name}' "
                    f"expects {self._format_type(expected_type)}, "
                    f"got {type(param_value).__name__}",
                    filename,
                    line_num,
                )

    def _check_type(self, value, expected_type):
        """Check if value matches expected type annotation.

        Handles:
        - Simple types (bool, int, str, float)
        - Optional[T] (Union[T, None])
        - Union[T1, T2, ...]
        - Literal[value1, value2, ...]

        Args:
            value: The value to check
            expected_type: The type annotation from signature

        Returns:
            True if type matches, False otherwise
        """
        # Handle None - only accept if type explicitly includes NoneType
        if value is None:
            origin = get_origin(expected_type)
            if origin is Union:
                args = get_args(expected_type)
                return type(None) in args
            return False

        # Handle Literal types (e.g., Literal[False], Literal["a", "b"])
        origin = get_origin(expected_type)
        if origin is Literal:
            literal_values = get_args(expected_type)
            return value in literal_values

        # Handle Union types (e.g., Union[str, int], Optional[str])
        if origin is Union:
            args = get_args(expected_type)
            # Check if value matches any of the union types
            for arg in args:
                if arg is type(None):
                    continue  # Skip NoneType
                # Recursively check each union member (handles nested Literal)
                if self._check_type(value, arg):
                    return True
            return False

        # Simple type check
        try:
            if isinstance(expected_type, type):
                return isinstance(value, expected_type)
        except TypeError:
            # Complex type annotation we can't check
            pass

        # Can't validate complex annotations, assume valid
        return True

    def _format_type(self, type_annotation):
        """Format type annotation for error messages.

        Converts type annotations to human-readable strings:
        - bool → "bool"
        - Optional[str] → "str or None"
        - Union[int, str] → "int or str"
        - Literal[False] → "False"
        - Literal["a", "b"] → "'a' or 'b'"

        Args:
            type_annotation: The type annotation to format

        Returns:
            Human-readable type string
        """
        origin = get_origin(type_annotation)

        # Handle Literal types
        if origin is Literal:
            literal_values = get_args(type_annotation)
            if len(literal_values) == 1:
                val = literal_values[0]
                return repr(val) if isinstance(val, str) else str(val)
            return " or ".join(
                repr(v) if isinstance(v, str) else str(v) for v in literal_values
            )

        if origin is Union:
            args = get_args(type_annotation)
            # Filter out NoneType for cleaner messages
            non_none = [arg for arg in args if arg is not type(None)]

            if len(non_none) == 1:
                # Optional[T] case - show as "T or None"
                if type(None) in args:
                    formatted = self._format_type(non_none[0])
                    return f"{formatted} or None"
                return self._format_type(non_none[0])

            # Union case - show all types
            type_names = []
            for arg in args:
                if arg is type(None):
                    type_names.append("None")
                else:
                    type_names.append(self._format_type(arg))
            return " or ".join(type_names)

        # Simple type
        if hasattr(type_annotation, "__name__"):
            return type_annotation.__name__

        # Fallback to string representation
        return str(type_annotation)

    @staticmethod
    def validate_fake(sv, context):
        """Validate fake StructuredValue calls and return a callable method.

        This is the validator for the StructuredValue syntax:
            fake: provider_name
        or with parameters:
            fake:
              - provider_name
              - param1
              - param2

        Args:
            sv: StructuredValue with function_name="fake"
            context: ValidationContext for error reporting

        Returns:
            A callable method that validates and executes Faker when called.
            The caller is responsible for calling the method (with args from sv if needed).
        """
        # Get provider name from first arg
        args = getattr(sv, "args", [])
        if not args:
            context.add_error(
                "fake: Missing provider name",
                getattr(sv, "filename", None),
                getattr(sv, "line_num", None),
            )
            return lambda *a, **kw: None

        provider_name = resolve_value(args[0], context)
        if not provider_name or not isinstance(provider_name, str):
            # Could not resolve provider name to a string
            return lambda *a, **kw: None

        # Check if Faker instance available
        if not context.faker_instance:
            # No Faker instance available - return None
            return lambda *a, **kw: None

        # Use FakerValidators to validate provider name
        validator = FakerValidators(context.faker_instance, context.faker_providers)

        # Validate provider name immediately
        resolved_name = validator.validate_provider_name(
            provider_name,
            context,
            getattr(sv, "filename", None),
            getattr(sv, "line_num", None),
        )

        if not resolved_name:
            # Validation failed, return mock placeholder
            return lambda *a, **kw: f"<fake_{provider_name}>"

        # Return a method that validates parameters and executes when called
        def validated_faker_method(*call_args, **call_kwargs):
            """Execute Faker method with parameter validation."""
            # Resolve parameters before validation and execution
            resolved_args = []
            for arg in call_args:
                resolved = resolve_value(arg, context)
                # Use resolved value if available, otherwise use original
                resolved_args.append(resolved if resolved is not None else arg)

            resolved_kwargs = {}
            for key, value in call_kwargs.items():
                resolved = resolve_value(value, context)
                resolved_kwargs[key] = resolved if resolved is not None else value

            # Validate parameters when method is called
            if resolved_args or resolved_kwargs:
                error_count_before = len(context.errors)
                validator.validate_provider_call(
                    resolved_name, resolved_args, resolved_kwargs, context
                )
                if len(context.errors) > error_count_before:
                    return f"<fake_{provider_name}>"

            # Execute Faker method with RESOLVED parameters
            try:
                method = getattr(context.faker_instance, resolved_name)
                return method(*resolved_args, **resolved_kwargs)
            except Exception as e:
                context.add_error(
                    f"fake.{resolved_name}: Execution error: {str(e)}",
                    getattr(sv, "filename", None),
                    getattr(sv, "line_num", None),
                )
                return f"<fake_{provider_name}>"

        return validated_faker_method



from base64 import b64encode
from snowfakery.plugins import SnowfakeryPlugin
from snowfakery.utils.validation_utils import resolve_value


class Base64(SnowfakeryPlugin):
    class Functions:
        def encode(self, data):
            return b64encode(bytes(str(data), "latin1")).decode("ascii")

    class Validators:
        """Validators for Base64 plugin functions."""

        @staticmethod
        def validate_encode(sv, context):
            """Validate Base64.encode(data)

            Args:
                sv: StructuredValue with args/kwargs
                context: ValidationContext for error reporting

            Returns:
                str: Base64-encoded mock data
            """

            kwargs = getattr(sv, "kwargs", {})
            args = getattr(sv, "args", [])

            # Check if data is provided (as positional or keyword argument)
            has_data = len(args) > 0 or "data" in kwargs

            if not has_data:
                context.add_error(
                    "Base64.encode: Missing required parameter 'data'",
                    getattr(sv, "filename", None),
                    getattr(sv, "line_num", None),
                )
                return b64encode(b"Mock data").decode("ascii")  # Fallback

            # WARNING: Unknown parameters (only 'data' is valid)
            valid_params = {"data"}
            unknown = set(kwargs.keys()) - valid_params
            if unknown:
                context.add_warning(
                    f"Base64.encode: Unknown parameter(s): {', '.join(sorted(unknown))}",
                    getattr(sv, "filename", None),
                    getattr(sv, "line_num", None),
                )

            # Try to resolve and encode the data
            data_val = args[0] if args else kwargs.get("data")
            resolved_data = resolve_value(data_val, context)

            if resolved_data is not None:
                try:
                    # Encode the resolved data
                    return b64encode(bytes(str(resolved_data), "latin1")).decode(
                        "ascii"
                    )
                except Exception:
                    pass

            return b64encode(b"Mock data").decode("ascii")  # Fallback
